"""Configuration file for the Sphinx documentation builder."""

import sys
import typing
from pathlib import Path

if "doctest" not in sys.argv:  # Avoid type checking/isinstance failures.
    # Tag used to avoid expanding arraylike alias in docs
    typing.GENERATING_DOCUMENTATION = True


sys.path.insert(0, Path("..").resolve())

project = "FlowJax"
copyright = "2022, Daniel Ward"
author = "Daniel Ward"

extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "nbsphinx",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "jax": ("https://jax.readthedocs.io/en/latest/", None),
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


html_theme = "sphinx_book_theme"
html_static_path = ["_static"]

html_css_files = ["style.css"]

html_theme_options = {
    "use_fullscreen_button": False,
    "use_download_button": False,
    "use_repository_button": True,
    "repository_url": "https://github.com/danielward27/flowjax",
    "home_page_in_toc": True,
    "logo": {
        "image_light": "_static/logo_light.svg",
        "image_dark": "_static/logo_dark.svg",
    },
}

html_title = "FlowJAX"
html_favicon = "_static/icon.svg"

pygments_style = "xcode"

copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True

napolean_use_rtype = False
napoleon_attr_annotations = True
napoleon_use_ivar = True

add_module_names = False
autodoc_inherit_docstrings = False
python_maximum_signature_line_length = 88



"""Planar transform.

A layer in a planar flow introduced in https://arxiv.org/pdf/1505.05770.pdf.
"""

from collections.abc import Callable
from functools import partial
from typing import ClassVar, Literal

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jax import nn
from jax.numpy.linalg import norm
from jaxtyping import Array, Float, PRNGKeyArray

from flowjax.bijections.bijection import AbstractBijection


class Planar(AbstractBijection):
    r"""Planar bijection as used by https://arxiv.org/pdf/1505.05770.pdf.

    Uses the transformation

    .. math::

        \boldsymbol{y}=\boldsymbol{x} +
            \boldsymbol{u} \cdot \text{tanh}(\boldsymbol{w}^T \boldsymbol{x} + b)

    where :math:`\boldsymbol{u} \in \mathbb{R}^D, \ \boldsymbol{w} \in \mathbb{R}^D`
    and :math:`b \in \mathbb{R}`. In the unconditional case, the (unbounded) parameters
    are learned directly. In the unconditional case they are parameterised by an MLP.

    Args:
        key: Jax random key.
        dim: Dimension of the bijection.
        cond_dim: Dimension of extra conditioning variables. Defaults to None.
        negative_slope: A positive float. If provided, then a leaky relu activation
            (with the corresponding negative slope) is used instead of tanh. This also
            provides the advantage that the bijection can be inverted analytically.
        **mlp_kwargs: Keyword arguments (excluding in_size and out_size) passed to
            the MLP (``equinox.nn.MLP``). Ignored when ``cond_dim`` is None.
    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None
    conditioner: Callable | None
    params: Array | None
    negative_slope: float | None

    def __init__(
        self,
        key: PRNGKeyArray,
        *,
        dim: int,
        cond_dim: int | None = None,
        negative_slope: float | None = None,
        **mlp_kwargs,
    ):
        self.shape = (dim,)

        if cond_dim is None:
            self.params = 0.01 * jr.normal(key, (2 * dim + 1,))
            self.conditioner = None
            self.cond_shape = None
        else:
            self.params = None
            self.conditioner = eqx.nn.MLP(cond_dim, 2 * dim + 1, **mlp_kwargs, key=key)
            self.cond_shape = (cond_dim,)

        self.negative_slope = negative_slope

    def transform(self, x, condition=None):
        return self.get_planar(condition).transform(x)

    def transform_and_log_det(self, x, condition=None):
        return self.get_planar(condition).transform_and_log_det(x)

    def inverse(self, y, condition=None):
        return self.get_planar(condition).inverse(y)

    def inverse_and_log_det(self, y, condition=None):
        return self.get_planar(condition).inverse_and_log_det(y)

    def get_planar(self, condition=None):
        """Get the planar bijection with the conditioning applied if conditional."""
        if self.cond_shape is not None:
            params = self.conditioner(condition)
        else:
            params = self.params
        dim = self.shape[0]
        assert params is not None
        w, u, bias = params[:dim], params[dim : 2 * dim], params[-1]
        return _UnconditionalPlanar(w, u, bias, self.negative_slope)


class _UnconditionalPlanar(AbstractBijection):
    """Unconditional planar bijection, used in Planar.

    Note act_scale (u in the paper) is unconstrained and the constraint to ensure
    invertiblitiy is applied in ``get_act_scale``.
    """

    shape: tuple[int, ...]
    cond_shape: ClassVar[None] = None
    weight: Array
    _act_scale: Array
    bias: Array
    activation: Literal["tanh"] | Literal["leaky_relu"]
    activation_fn: Callable
    negative_slope: float | None

    def __init__(
        self,
        weight: Float[Array, " dim"],
        act_scale: Float[Array, " dim"],
        bias: Float[Array, " "],
        negative_slope: float | None = None,
    ):
        self.weight = weight
        self.bias = bias
        self.shape = weight.shape
        self.negative_slope = negative_slope
        self._act_scale = act_scale

        if negative_slope is None:
            self.activation = "tanh"
            self.activation_fn = jnp.tanh
        else:
            if negative_slope <= 0:
                raise ValueError("The negative slope value should be >0.")
            self.activation = "leaky_relu"
            self.activation_fn = partial(nn.leaky_relu, negative_slope=negative_slope)

    def transform(self, x, condition=None):
        u = self.get_act_scale()
        return x + u * self.activation_fn(self.weight @ x + self.bias)

    def transform_and_log_det(self, x, condition=None):
        u = self.get_act_scale()
        act = self.activation_fn(x @ self.weight + self.bias)
        y = x + u * act
        if self.activation == "leaky_relu":
            psi = jnp.where(act < 0, self.negative_slope, 1) * self.weight
        else:
            psi = (1 - act**2) * self.weight
        log_det = jnp.log(jnp.abs(1 + u @ psi))
        return y, log_det

    def get_act_scale(self):
        """Apply constraint to u to ensure invertibility.

        See appendix A1 in https://arxiv.org/pdf/1505.05770.pdf.
        """
        wtu = self._act_scale @ self.weight
        m_wtu = -1 + jnp.log(1 + nn.softplus(wtu))
        return self._act_scale + (m_wtu - wtu) * self.weight / norm(self.weight) ** 2

    def inverse(self, y, condition=None):
        if self.activation != "leaky_relu":
            raise NotImplementedError(
                "The inverse planar transformation is only implemented with the leaky "
                "relu activation function.",
            )
        return self.inverse_and_log_det(y, condition)[0]

    def inverse_and_log_det(self, y, condition=None):
        if self.activation != "leaky_relu":
            raise NotImplementedError(
                "The inverse planar transformation is only implemented with the leaky "
                "relu activation function.",
            )
        # Expanding explanation as the inverse is not in the original paper.
        # The derivation steps for the inversion are:
        # 1. Let z = w^Tx+b
        # 2. We want x=y-uσ(z), where σ is the leaky relu function.
        # 3. Sub x=y-uσ(z) into z = w^Tx+b,
        # 4. Solve for z, which gives z = (w^Ty+b)/(1+w^Tus), where s is the slope
        #   σ'(z), i.e. s=1 if z>=0 and s=negative_slope otherwise. To find the
        #   slope, it is sufficient to check the sign of the numerator w^Ty+b, rather
        #   than z, as the denominator is constrained to be positive.
        # 5. Compute inverse using x=y-uσ(z)

        numerator = self.weight @ y + self.bias
        relu_slope = jnp.where(numerator < 0, self.negative_slope, 1)
        us = self.get_act_scale() * relu_slope
        denominator = 1 + self.weight @ us
        log_det = -jnp.log(jnp.abs(1 + us @ self.weight))
        x = y - us * (numerator / denominator)
        return x, log_det



"""Power transform."""

from typing import ClassVar

import equinox as eqx
import jax.numpy as jnp

from flowjax.bijections.bijection import AbstractBijection


class Power(AbstractBijection):
    """Power transform :math:`y = x^p`.

    Supports postive values, over which this is a bijection.

    Args:
        exponent: The exponent.
        shape: The shape of the bijection.
    """

    exponent: int | float
    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def transform(self, x, condition=None):
        x = eqx.error_if(x, x < 0, "Negative values not supported for Power.")
        return x**self.exponent

    def inverse(self, y, condition=None):
        y = eqx.error_if(y, y < 0, "Negative values not supported for Power.")
        return y ** (1 / self.exponent)

    def transform_and_log_det(self, x, condition=None):
        y = self.transform(x)
        return y, jnp.log(jnp.abs(self.exponent * y / x)).sum()

    def inverse_and_log_det(self, y, condition=None):
        x = self.inverse(y)
        return x, -jnp.log(jnp.abs(self.exponent * y / x)).sum()



"""Sigmoid bijection."""

from typing import ClassVar

import jax.numpy as jnp
from jax import nn
from jax.scipy.special import logit

from flowjax.bijections.bijection import AbstractBijection


class Sigmoid(AbstractBijection):
    r"""Sigmoid bijection :math:`y = \sigma(x) = \frac{1}{1 + \exp(-x)}`.

    Args:
        shape: The shape of the transform.
    """

    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def transform(self, x, condition=None):
        return nn.sigmoid(x)

    def transform_and_log_det(self, x, condition=None):
        y = nn.sigmoid(x)
        log_det = jnp.sum(nn.log_sigmoid(x) + nn.log_sigmoid(-x))
        return y, log_det

    def inverse(self, y, condition=None):
        return logit(y)

    def inverse_and_log_det(self, y, condition=None):
        x = logit(y)
        log_det = -jnp.sum(nn.log_sigmoid(x) + nn.log_sigmoid(-x))
        return x, log_det



""":class:`AbstractUnwrappable` objects and utilities.

These are "placeholder" values for specifying custom behaviour for nodes in a pytree.
Many of these facilitate similar functions to pytorch parameterizations. We use this
for example to apply parameter constraints, masking of parameters etc. To apply the
behaviour, we use :func:`unwrap`, which will replace any :class:`AbstractUnwrappable`
nodes in a pytree with the unwrapped versions.

Unwrapping is automatically called in several places, primarily:

* Prior to calling the bijection methods: ``transform``, ``inverse``,
  ``transform_and_log_det`` and ``inverse_and_log_det``.
* Prior to calling distribution methods: ``log_prob``, ``sample`` and
  ``sample_and_log_prob``.
* Prior to computing the loss functions.


.. note::
    
    If creating a custom unwrappable, remember that unwrapping will generally occur
    after initialization of the model. Because of this, we recommend ensuring that
    the ``unwrap`` method supports unwrapping if the model is constructed in a
    vectorized context, such as ``eqx.filter_vmap``, e.g. through broadcasting or
    vectorization.
"""

from abc import abstractmethod
from collections.abc import Callable, Iterable
from typing import Any, Generic, TypeVar

import equinox as eqx
import jax
import jax.numpy as jnp
from jax import lax
from jax.nn import softplus
from jaxtyping import Array, Int, PyTree, Scalar

from flowjax.utils import inv_softplus

T = TypeVar("T")


def unwrap(tree: PyTree):
    """Recursively unwraps all :class:`AbstractUnwrappable` nodes within a pytree.

    This leaves all other nodes unchanged. If nested, the innermost
    ``AbstractUnwrappable`` nodes are unwrapped first.

    Example:
        Enforcing positivity.

        .. doctest::

            >>> from flowjax.wrappers import Parameterize, unwrap
            >>> import jax.numpy as jnp
            >>> params = Parameterize(jnp.exp, jnp.zeros(3))
            >>> unwrap(("abc", 1, params))
            ('abc', 1, Array([1., 1., 1.], dtype=float32))
    """

    def _map_fn(leaf):
        if isinstance(leaf, AbstractUnwrappable):
            # Flatten to ignore until all contained AbstractUnwrappables are unwrapped
            flat, tree_def = eqx.tree_flatten_one_level(leaf)
            tree = jax.tree_util.tree_unflatten(tree_def, unwrap(flat))
            return tree.unwrap()
        return leaf

    return jax.tree_util.tree_map(
        f=_map_fn,
        tree=tree,
        is_leaf=lambda x: isinstance(x, AbstractUnwrappable),
    )


class AbstractUnwrappable(eqx.Module, Generic[T]):
    """An abstract class representing an unwrappable object.

    Unwrappables generally replace nodes in a pytree, in order to specify some custom
    behaviour to apply upon unwrapping before use. This can be used e.g. to apply
    parameter constraints, such as making scale parameters postive, or applying
    stop_gradient before accessing the parameters.
    """

    @abstractmethod
    def unwrap(self) -> T:
        """Returns the unwrapped pytree, assuming no wrapped subnodes exist."""
        pass


class Parameterize(AbstractUnwrappable[T]):
    """Unwrap an object by calling fn with args and kwargs.

    All of fn, args and kwargs may contain trainable parameters. If the Parameterize is
    created within ``eqx.filter_vmap``, unwrapping is automatically vectorized
    correctly, as long as the vmapped constructor adds leading batch
    dimensions to all arrays (the default for ``eqx.filter_vmap``).

    Example:
        .. doctest::

            >>> from flowjax.wrappers import Parameterize, unwrap
            >>> import jax.numpy as jnp
            >>> positive = Parameterize(jnp.exp, jnp.zeros(3))
            >>> unwrap(positive)  # Aplies exp on unwrapping
            Array([1., 1., 1.], dtype=float32)

    Args:
        fn: Callable to call with args, and kwargs.
        *args: Positional arguments to pass to fn.
        **kwargs: Keyword arguments to pass to fn.
    """

    fn: Callable[..., T]
    args: Iterable
    kwargs: dict[str, Any]
    _dummy: Int[Scalar, ""]  # Used to track vectorized construction.

    def __init__(self, fn: Callable, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self._dummy = jnp.empty((), int)

    def unwrap(self) -> T:

        def _unwrap_fn(self):
            return self.fn(*self.args, **self.kwargs)

        for dim in reversed(self._dummy.shape):  # vectorize if constructed under vmap
            _unwrap_fn = eqx.filter_vmap(_unwrap_fn, axis_size=dim)
        return _unwrap_fn(self)


class NonTrainable(AbstractUnwrappable[T]):
    """Applies stop gradient to all arraylike leaves before unwrapping.

    See also :func:`non_trainable`, which is probably a generally prefereable way to
    achieve similar behaviour, which wraps the arraylike leaves directly, rather than
    the tree. Useful to mark pytrees (arrays, submodules, etc) as frozen/non-trainable.
    We also filter out NonTrainable nodes when partitioning parameters for training,
    or when parameterizing bijections in coupling/masked autoregressive flows
    (transformers).
    """

    tree: T

    def unwrap(self) -> T:
        differentiable, static = eqx.partition(self.tree, eqx.is_array_like)
        return eqx.combine(lax.stop_gradient(differentiable), static)


def non_trainable(tree: PyTree):
    """Freezes parameters by wrapping inexact array leaves with :class:`NonTrainable`.

    .. note::

        Regularization is likely to apply before unwrapping. To avoid regularization
        impacting non-trainable parameters, they should be filtered out,
        for example using:

        .. code-block:: python

            >>> eqx.partition(
            ...     ...,
            ...     is_leaf=lambda leaf: isinstance(leaf, wrappers.NonTrainable),
            ... )


        This is done in both :func:`~flowjax.train.fit_to_data` and
        :func:`~flowjax.train.fit_to_variational_target`.

    Wrapping the arrays rather than the entire tree is often preferable, allowing easier
    access to attributes compared to wrapping the entire tree.

    Args:
        tree: The pytree.
    """

    def _map_fn(leaf):
        return NonTrainable(leaf) if eqx.is_inexact_array(leaf) else leaf

    return jax.tree_util.tree_map(
        f=_map_fn,
        tree=tree,
        is_leaf=lambda x: isinstance(x, NonTrainable),
    )


class WeightNormalization(AbstractUnwrappable[Array]):
    """Applies weight normalization (https://arxiv.org/abs/1602.07868).

    Args:
        weight: The (possibly wrapped) weight matrix.
    """

    weight: Array | AbstractUnwrappable[Array]
    scale: Array | AbstractUnwrappable[Array] = eqx.field(init=False)

    def __init__(self, weight: Array | AbstractUnwrappable[Array]):
        self.weight = weight
        scale_init = 1 / jnp.linalg.norm(unwrap(weight), axis=-1, keepdims=True)
        self.scale = Parameterize(softplus, inv_softplus(scale_init))

    def unwrap(self) -> Array:
        weight_norms = jnp.linalg.norm(self.weight, axis=-1, keepdims=True)
        return self.scale * self.weight / weight_norms



import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import pytest
from jax import vmap
from jax.tree_util import tree_map

from flowjax.bijections import RationalQuadraticSpline
from flowjax.bijections.rational_quadratic_spline import _real_to_increasing_on_interval


@pytest.mark.parametrize("interval", [3, (-4, 5)])
def test_RationalQuadraticSpline_tails(interval):
    key = jr.key(0)
    spline = RationalQuadraticSpline(knots=10, interval=interval)

    # Change to random initialisation, rather than identity.
    spline = tree_map(
        lambda x: jr.normal(key, x.shape) if eqx.is_inexact_array(x) else x,
        spline,
    )

    x = jr.uniform(key, (5,), minval=spline.interval[0], maxval=spline.interval[1])
    y = vmap(spline.transform)(x)
    assert pytest.approx(x, abs=1e-5) != y

    # Outside interval, default to identity
    x = jnp.array([spline.interval[0] - 1, spline.interval[1] + 1])
    y = vmap(spline.transform)(x)
    assert pytest.approx(x, abs=1e-5) == y


@pytest.mark.parametrize("interval", [3, (-4, 5)])
def test_RationalQuadraticSpline_init(interval):
    # Test it is initialized at the identity
    x = jnp.array([-7, 0.1, 2, 1])
    spline = RationalQuadraticSpline(knots=10, interval=interval)
    y = vmap(spline.transform)(x)
    assert pytest.approx(x, abs=1e-6) == y


def test_real_to_increasing_on_interval():
    y = _real_to_increasing_on_interval(jnp.array([-3.0, -4, 5, 0, 0]), (-3, 7))
    assert y.max() == 7
    assert y.min() == -3
    assert jnp.all(jnp.diff(y)) > 0



from math import prod

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
import pytest
from jax.tree_util import tree_map

from flowjax.distributions import Normal
from flowjax.wrappers import (
    NonTrainable,
    Parameterize,
    WeightNormalization,
    non_trainable,
    unwrap,
)


def test_Parameterize():
    diag = Parameterize(jnp.diag, jnp.ones(3))
    assert pytest.approx(jnp.eye(3)) == unwrap(diag)


def test_nested_Parameterized():
    param = Parameterize(
        jnp.square,
        Parameterize(jnp.square, Parameterize(jnp.square, 2)),
    )
    assert unwrap(param) == jnp.square(jnp.square(jnp.square(2)))


def test_non_trainable():
    dist = non_trainable(Normal())

    def loss(dist, x):
        return dist.log_prob(x)

    grad = eqx.filter_grad(loss)(dist, 1)
    assert pytest.approx(0) == jax.flatten_util.ravel_pytree(grad)[0]


def test_WeightNormalization():
    arr = jr.normal(jr.key(0), (10, 3))
    weight_norm = WeightNormalization(arr)

    # Unwrapped norms should match weightnorm scale parameter
    expected = unwrap(weight_norm.scale)
    assert pytest.approx(expected) == jnp.linalg.norm(
        unwrap(weight_norm), axis=-1, keepdims=True
    )


test_cases = {
    "NonTrainable": lambda key: NonTrainable(jr.normal(key, 10)),
    "Parameterize-exp": lambda key: Parameterize(jnp.exp, jr.normal(key, 10)),
    "Parameterize-diag": lambda key: Parameterize(jnp.diag, jr.normal(key, 10)),
    "WeightNormalization": lambda key: WeightNormalization(jr.normal(key, (10, 2))),
}


@pytest.mark.parametrize("shape", [(), (2,), (5, 2, 4)])
@pytest.mark.parametrize("wrapper_fn", test_cases.values(), ids=test_cases.keys())
def test_vectorization_invariance(wrapper_fn, shape):
    keys = jr.split(jr.key(0), prod(shape))
    wrapper = wrapper_fn(keys[0])  # Standard init

    # Multiple vmap init - should have same result in zero-th index
    vmap_wrapper_fn = wrapper_fn
    for _ in shape:
        vmap_wrapper_fn = eqx.filter_vmap(vmap_wrapper_fn)

    vmap_wrapper = vmap_wrapper_fn(keys.reshape(shape))

    unwrapped = unwrap(wrapper)
    unwrapped_vmap = unwrap(vmap_wrapper)
    unwrapped_vmap_zero = tree_map(
        lambda leaf: leaf[*([0] * len(shape)), ...],
        unwrapped_vmap,
    )
    assert eqx.tree_equal(unwrapped, unwrapped_vmap_zero, atol=1e-7)



import jax.numpy as jnp
import jax.random as jr

from flowjax.train.losses import _get_contrastive_idxs


def test_get_contrastive_idxs():
    key = jr.key(0)
    batch_size = 5

    for _ in range(5):
        key, subkey = jr.split(key)
        idxs = _get_contrastive_idxs(subkey, batch_size=batch_size, n_contrastive=4)
        for i, row in enumerate(idxs):
            assert i not in row

        assert jnp.all(idxs < batch_size)
